#!/usr/bin/env python
# vim: set fileencoding=utf-8 :

###############################################################################
#                                                                             #
# Copyright (c) 2017 Idiap Research Institute, http://www.idiap.ch/           #
# Contact: beat.support@idiap.ch                                              #
#                                                                             #
# This file is part of the beat.web module of the BEAT platform.              #
#                                                                             #
# Commercial License Usage                                                    #
# Licensees holding valid commercial BEAT licenses may use this file in       #
# accordance with the terms contained in a written agreement between you      #
# and Idiap. For further information contact tto@idiap.ch                     #
#                                                                             #
# Alternatively, this file may be used under the terms of the GNU Affero      #
# Public License version 3 as published by the Free Software and appearing    #
# in the file LICENSE.AGPL included in the packaging of this file.            #
# The BEAT platform is distributed in the hope that it will be useful, but    #
# WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY  #
# or FITNESS FOR A PARTICULAR PURPOSE.                                        #
#                                                                             #
# You should have received a copy of the GNU Affero Public License along      #
# with the BEAT platform. If not, see http://www.gnu.org/licenses/.           #
#                                                                             #
###############################################################################

import datetime

import logging
logger = logging.getLogger(__name__)

import simplejson

from django.db import utils
from django.db import models
from django.conf import settings

import beat.core.data
import beat.core.hash

from .result import Result


#----------------------------------------------------------


class JobManager(models.Manager):

    def create_job(self, block):

        # Compute the key of the job
        hashes = [ x.hash for x in block.outputs.order_by('hash') ]
        key = beat.core.hash.hash(''.join(hashes))

        # Determine if the job can be run or is dependent on others
        runnable_date = None
        if block.is_runnable():
            runnable_date = datetime.datetime.now()

        # Create the job
        job = self.model(
            block=block,
            key=key,
            runnable_date=runnable_date
        )

        job.save()

        return job


#----------------------------------------------------------


class Job(models.Model):
    '''Class describing the execution of a Job on the backend'''

    block = models.OneToOneField('experiments.Block', null=True,
                                 on_delete=models.CASCADE, related_name='job')

    result = models.OneToOneField(Result, null=True, on_delete=models.CASCADE,
                                  related_name='job')

    runnable_date = models.DateTimeField(null=True, blank=True)

    start_date = models.DateTimeField(null=True, blank=True)

    end_date = models.DateTimeField(null=True, blank=True)

    key = models.CharField(max_length=64)

    mirror = models.BooleanField(default=False)


    objects = JobManager()


    def __str__(self):
        return "Job(%s, %s, key=%s, mirror=%s, splits=%d, cores=%d)" % \
            (self.block.name, self.block.experiment.name,
             self.key, str(self.mirror),
             self.block.required_slots,
             self.block.queue.cores_per_slot)


#----------------------------------------------------------


class JobSplitManager(models.Manager):

    def create_splits(self, job):

        # If no splitting is required, only create one split
        if job.block.required_slots == 1:
            split = JobSplit(job=job, split_index=0)
            split.save()
            return [split]


        # Retrieve the list of synchronized inputs
        configuration = simplejson.loads(job.block.command)
        inputs = [ entry for name, entry in configuration['inputs'].items()
                   if entry['channel'] == configuration['channel'] ]


        # Load the list of indices for each inputs
        indices = []
        for input_cfg in inputs:
            if 'database' in input_cfg:
                indices.extend(beat.core.data.load_data_index_db(settings.CACHE_ROOT, input_cfg['path']))
            else:
                indices.append(beat.core.data.load_data_index(settings.CACHE_ROOT, input_cfg['path']))


        # Attempt to split the indices
        nb_splits = job.block.required_slots
        while nb_splits > 1:

            # Determine N splits using the possible indices for split:
            split_indices = beat.core.data.foundSplitRanges(indices, nb_splits)

            if len(split_indices) > 0:
                break

            nb_splits -= 1

        if nb_splits != job.block.required_slots:
           message = "The processing of the block `%s' of experiment `%s' " \
                     "was splitted in %d instead of the requested %d" % \
                     (job.block.name, job.block.experiment.fullname(),
                      nb_splits, job.block.required_slots)
           logger.warning(message)


        # Create the necessary splits and assign the ranges
        splits = []
        for i, indices in enumerate(split_indices):
            split = JobSplit(
                job=job,
                split_index=i,
                start_index=indices[0],
                end_index=indices[1],
            )
            split.save()

            splits.append(split)

        return splits


#----------------------------------------------------------


class JobSplit(models.Model):
    '''Class describing a part of job of an experiment'''

    QUEUED     = 'N'
    PROCESSING = 'P'
    COMPLETED  = 'C'
    FAILED     = 'F'
    CANCELLED  = 'L'
    CANCELLING = 'K'

    STATUS = (
        (QUEUED,     'Queued'),
        (PROCESSING, 'Processing'),
        (COMPLETED,  'Completed'),
        (FAILED,     'Failed'),
        (CANCELLED,  'Cancelled'),
        (CANCELLING, 'Cancelling'),
    )


    worker = models.ForeignKey('Worker', null=True, on_delete=models.SET_NULL,
                               related_name='splits')

    job = models.ForeignKey(Job, null=True, on_delete=models.CASCADE,
                            related_name='splits')

    split_index = models.PositiveIntegerField()

    start_index = models.PositiveIntegerField(null=True)

    end_index = models.PositiveIntegerField(null=True)

    status = models.CharField(max_length=1, choices=STATUS, default=QUEUED)

    result = models.OneToOneField(Result, null=True, on_delete=models.CASCADE,
                                  related_name='split')

    start_date = models.DateTimeField(null=True)

    end_date = models.DateTimeField(null=True)


    objects = JobSplitManager()


    class Meta:
        unique_together = ('job', 'split_index')


    def __str__(self):
        return "JobSplit(%s, index=%d, state=%s)%s" % \
            (self.job, self.split_index, self.status,
                ('@%s' % self.worker) if self.worker else '')


    def done(self):
        '''Says whether the job has finished or not'''

        return self.status in (
            JobSplit.COMPLETED,
            JobSplit.FAILED,
            JobSplit.CANCELLED,
        )
