#!/usr/bin/env python
# vim: set fileencoding=utf-8 :

###############################################################################
#                                                                             #
# Copyright (c) 2017 Idiap Research Institute, http://www.idiap.ch/           #
# Contact: beat.support@idiap.ch                                              #
#                                                                             #
# This file is part of the beat.web module of the BEAT platform.              #
#                                                                             #
# Commercial License Usage                                                    #
# Licensees holding valid commercial BEAT licenses may use this file in       #
# accordance with the terms contained in a written agreement between you      #
# and Idiap. For further information contact tto@idiap.ch                     #
#                                                                             #
# Alternatively, this file may be used under the terms of the GNU Affero      #
# Public License version 3 as published by the Free Software and appearing    #
# in the file LICENSE.AGPL included in the packaging of this file.            #
# The BEAT platform is distributed in the hope that it will be useful, but    #
# WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY  #
# or FITNESS FOR A PARTICULAR PURPOSE.                                        #
#                                                                             #
# You should have received a copy of the GNU Affero Public License along      #
# with the BEAT platform. If not, see http://www.gnu.org/licenses/.           #
#                                                                             #
###############################################################################

from django.test import TestCase
from django.conf import settings

from ...utils.management.commands import install
from ...experiments.models import Experiment
from ...experiments.models import Block
from ...experiments.models import CachedFile
from ...algorithms.models import Algorithm

from ..models import Queue
from ..models import Worker
from ..models import Environment
from ..models import Job
from ..models import JobSplit

from ..utils import setup_backend
from ..management.commands import qsetup

from beat.core.dataformat import DataFormat
from beat.core.data import CachedDataSink
from beat.core.database import Database
import beat.core.hash

import os
import sys


#----------------------------------------------------------


ONE_QUEUE_TWO_WORKERS = {
    "queues": {
        "queue": {
            "memory-limit": 4*1024,
            "time-limit": 1440, #1 day
            "cores-per-slot": 1,
            "max-slots-per-user": 2,
            "environments": [
                'Python 2.7 (1.3.0)'
            ],
            "slots": {
                'node1': {
                    "quantity": 1,
                    "priority": 0
                },
                'node2': {
                    "quantity": 1,
                    "priority": 0
                }
            },
            "groups": [
                "Default",
            ],
        }
    },
    "workers": {
        "node1": {
            "cores": 1,
            "memory": 16*1024,
        },
        "node2": {
            "cores": 1,
            "memory": 16*1024,
        }
    },
    "environments": {
        "Python 2.7 (1.3.0)": {
            "name": 'Python 2.7',
            "version": '1.3.0',
            "short_description": "Test",
            "description": "Test environment",
            "languages": ["python"],
        },
    },
}


#----------------------------------------------------------


class BackendUtilitiesMixin(object):

    @classmethod
    def setup_test_data(cls):
        install.create_sites()
        system_user, plot_user, user = install.create_users('user', 'user')
        install.add_group('Default')

        setup_backend(qsetup.DEFAULT_CONFIGURATION)

        Worker.objects.update(active=True)
        env = Environment.objects.get(name='Python 2.7')
        queue = Queue.objects.first()

        template_data = dict(
            system_user = system_user,
            plot_user = plot_user,
            user = user,
            private = False,
            queue = queue.name,
            environment = dict(name=env.name, version=env.version),
        )

        prefix = os.path.join(
            os.path.dirname(os.path.dirname(os.path.realpath(sys.argv[0]))),
            'src',
            'beat.examples',
        )

        install.install_contributions(prefix, 'system', template_data)
        install.install_contributions(prefix, 'test', template_data)

        if not os.path.exists(settings.CACHE_ROOT):
          os.mkdir(settings.CACHE_ROOT)


    def clean_cache(self):
        for p, dirs, files in os.walk(settings.CACHE_ROOT, topdown=False):

            files = [f for f in files if not f.startswith('.')]
            dirs[:] = [d for d in dirs if not d.startswith('.')] #note: in-place

            for f in files:
                fullpath = os.path.join(p, f)
                os.remove(fullpath)

            for d in dirs:
                fullpath = os.path.join(p, d)
                os.rmdir(fullpath)


    def set_experiment_state(self, experiment, experiment_status=None, block_status=None,
                             cache_status=None):
        if block_status:
            for name, status in block_status.items():
                block = experiment.blocks.get(name=name)
                block.status = status
                block.save()

        if cache_status:
            for name, status in cache_status.items():
                block = experiment.blocks.get(name=name)
                for cached_file in block.outputs.all():
                    cached_file.status = status
                    cached_file.save()

        if experiment_status:
            experiment.status = experiment_status
            experiment.save()


    def generate_cached_files(self, hash, splits):
        dataformat = DataFormat(settings.PREFIX, 'system/integer/1')

        path = os.path.join(settings.CACHE_ROOT, beat.core.hash.toPath(hash))
        os.makedirs(os.path.dirname(path))

        value = 0

        for index, split in enumerate(splits):
            sink = CachedDataSink()

            start_data_index = split[0][0] if isinstance(split[0], tuple) else split[0]
            end_data_index = split[-1][1] if isinstance(split[-1], tuple) else split[-1]

            sink.setup(path, dataformat, start_data_index, end_data_index)

            for indices in split:
                if not isinstance(indices, tuple):
                    start = indices
                    end = indices
                else:
                    start = indices[0]
                    end = indices[1]

                sink.write({
                        'value': value,
                    },
                    start_data_index = start,
                    end_data_index = end
                )

                value += 1

            sink.close()


    def prepare_databases(self, configuration):
        for _, cfg in configuration['datasets'].items():
            path = beat.core.hash.toPath(beat.core.hash.hashDataset(
                    cfg['database'], cfg['protocol'], cfg['set']), suffix='.db')

            if not os.path.exists(os.path.join(settings.CACHE_ROOT, path)):
                database = Database(settings.PREFIX, cfg['database'])
                view = database.view(cfg['protocol'], cfg['set'])
                view.index(os.path.join(settings.CACHE_ROOT, path))


#----------------------------------------------------------


class BaseBackendTestCase(TestCase, BackendUtilitiesMixin):

    @classmethod
    def setUpTestData(cls):
      cls.setup_test_data()


    def setUp(self):
        self.clean_cache()


    def tearDown(self):
        self.clean_cache()


    def check_single(self, xp):
        '''Checks user/user/single/1/single'''

        self.assertEqual(xp.blocks.count(), 2)

        b0 = xp.blocks.all()[0]

        self.assertEqual(b0.name, 'echo')
        self.assertEqual(b0.status, Block.PENDING)
        self.assertEqual(b0.algorithm, Algorithm.objects.get(name='integers_echo'))
        self.assertEqual(b0.dependencies.count(), 0)
        self.assertEqual(b0.dependents.count(), 1)
        self.assertEqual(b0.queue.name, 'queue')
        self.assertEqual(b0.environment.name, 'Python 2.7')
        self.assertEqual(b0.required_slots, 1)
        self.assertEqual(b0.inputs.count(), 1)
        self.assertEqual(b0.outputs.count(), 1)
        self.assertEqual(b0.job.splits.count(), 0) #not scheduled yet

        assert not b0.done()

        b1 = xp.blocks.all()[1]

        self.assertEqual(b1.name, 'analysis')
        self.assertEqual(b1.status, Block.PENDING)
        self.assertEqual(b1.algorithm, Algorithm.objects.get(name='integers_echo_analyzer'))
        self.assertEqual(b1.dependencies.count(), 1)
        self.assertEqual(b1.dependents.count(), 0)
        self.assertEqual(b1.queue.name, 'queue')
        self.assertEqual(b1.environment.name, 'Python 2.7')
        self.assertEqual(b1.required_slots, 1)
        self.assertEqual(b1.inputs.count(), 1)
        self.assertEqual(b1.outputs.count(), 1)
        self.assertEqual(b1.job.splits.count(), 0) #not scheduled yet

        assert not b1.done()
