#!/usr/bin/env python
# vim: set fileencoding=utf-8 :

###############################################################################
#                                                                             #
# Copyright (c) 2017 Idiap Research Institute, http://www.idiap.ch/           #
# Contact: beat.support@idiap.ch                                              #
#                                                                             #
# This file is part of the beat.web module of the BEAT platform.              #
#                                                                             #
# Commercial License Usage                                                    #
# Licensees holding valid commercial BEAT licenses may use this file in       #
# accordance with the terms contained in a written agreement between you      #
# and Idiap. For further information contact tto@idiap.ch                     #
#                                                                             #
# Alternatively, this file may be used under the terms of the GNU Affero      #
# Public License version 3 as published by the Free Software and appearing    #
# in the file LICENSE.AGPL included in the packaging of this file.            #
# The BEAT platform is distributed in the hope that it will be useful, but    #
# WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY  #
# or FITNESS FOR A PARTICULAR PURPOSE.                                        #
#                                                                             #
# You should have received a copy of the GNU Affero Public License along      #
# with the BEAT platform. If not, see http://www.gnu.org/licenses/.           #
#                                                                             #
###############################################################################

import time
import collections

from django.core import management

from .common import BaseBackendTestCase

from ...experiments.models import Experiment

from ..models import Queue
from ..models import Worker
from ..models import Slot
from ..utils import dump_backend
from ..utils import setup_backend
from ..management.commands import qsetup



# Example configuration with 3 queues with an increasing amount of resources
# running on the same host
QUEUES_WITHOUT_PRIORITY = {
    "queues": collections.OrderedDict([
        ("q1", {
            "memory-limit": 4*1024,
          "time-limit": 180, #3 hours
          "cores-per-slot": 1,
          "max-slots-per-user": 4,
          "environments": ['Python 2.7 (1.3.0)'],
          "groups": [
              "Default",
          ],
          "slots": {
              "node1": {
                  "quantity": 4,
                  "priority": 0
              }
          }
        }
        ),
        ("q2", {
            "memory-limit": 8*1024,
            "time-limit": 360, #6 hours
            "cores-per-slot": 2,
            "max-slots-per-user": 2,
            "environments": ['Python 2.7 (1.3.0)'],
            "groups": [
                "Default",
            ],
            "slots": {
                "node1": {
                    "quantity": 2,
                    "priority": 0
                },
            }
        }
        ),
        ("q4", {
            "memory-limit": 16*1024,
            "time-limit": 720, #12 hours
            "cores-per-slot": 4,
            "max-slots-per-user": 1,
            "environments": ['Python 2.7 (1.3.0)'],
            "groups": [
                "Default",
            ],
            "slots": {
                "node1": {
                    "quantity": 1,
                    "priority": 0
                },
            }
        }
        )
    ]),
  "workers": {
      "node1": {
          "cores": 4,
          "memory": 16*1024,
      }
    },
  "environments": {
      'Python 2.7 (1.3.0)': {
          "name": 'Python 2.7',
          "version": '1.3.0',
          "short_description": "Test",
          "description": "Test environment",
          "languages": "python",
      },
    },
}

# Example configuration with 3 queues sharing slots on 2 hosts
PRIORITY_QUEUES = {
    "queues": collections.OrderedDict([
        ("q1", {
            "memory-limit": 4*1024,
            "time-limit": 180, #3 hours
            "cores-per-slot": 1,
            "max-slots-per-user": 2,
            "environments": ['Python 2.7 (1.3.0)'],
            "groups": [
                "Default",
            ],
            "slots": {
                "node1": {
                    "quantity": 4,
                    "priority": 5
                },
                "node2": {
                    "quantity": 4,
                    "priority": 0
                },
            }
        },
        ),
      ("q2", {
          "memory-limit": 8*1024,
        "time-limit": 360, #6 hours
        "cores-per-slot": 2,
        "max-slots-per-user": 1,
        "environments": ['Python 2.7 (1.3.0)'],
        "groups": [
            "Default",
        ],
        "slots": {
            "node1": {
                "quantity": 2,
                "priority": 0
            },
          "node2": {
              "quantity": 2,
                "priority": 10
            }
        }
      },
        ),
      ("q1_special", {
          "memory-limit": 4*1024,
        "time-limit": 180, #3 hours
        "cores-per-slot": 1,
        "max-slots-per-user": 8,
        "environments": ['Python 2.7 (1.3.0)'],
        "groups": [
            "Default",
        ],
        "slots": {
            "node1": {
                "quantity": 4,
                "priority": 0
            },
          "node2": {
              "quantity": 4,
                "priority": 5
            }
        }
      }
        ),
    ]),
    "workers": collections.OrderedDict([
        ("node1", {
            "cores": 4,
            "memory": 32*1024,
        }
        ),
      ("node2", {
          "cores": 4,
        "memory": 16*1024,
      }
        )
    ]),
    "environments": {
        'Python 2.7 (1.3.0)': {
            "name": 'Python 2.7',
          "version": '1.3.0',
          "short_description": "Test",
          "description": "Test environment",
          "languages": "python",
        },
    },
}



class BackendSetup(BaseBackendTestCase):

    def check_default_config(self):

        # checks all is there
        dumped_config = dump_backend()

        self.assertEqual(dumped_config['workers'], qsetup.DEFAULT_CONFIGURATION['workers'])
        self.assertEqual(dumped_config['environments'], qsetup.DEFAULT_CONFIGURATION['environments'])
        self.assertEqual(sorted(dumped_config['queues'].keys()),
                         sorted(qsetup.DEFAULT_CONFIGURATION['queues'].keys()))

        for queue in dumped_config['queues'].keys():
            self.assertEqual(sorted(dumped_config['queues'][queue].keys()),
                             sorted(qsetup.DEFAULT_CONFIGURATION['queues'][queue].keys()))

            for key in dumped_config['queues'][queue].keys():
                if key == 'environments':
                    continue

                self.assertEqual(dumped_config['queues'][queue][key],
                                 qsetup.DEFAULT_CONFIGURATION['queues'][queue][key])

            self.assertEqual(sorted(dumped_config['queues'][queue]['environments']),
                             sorted(qsetup.DEFAULT_CONFIGURATION['queues'][queue]['environments']))

        worker = Worker.objects.get()
        queue = Queue.objects.get()
        Worker.objects.update(active=True)

        self.assertEqual(worker.available_cores(), qsetup.CORES)
        self.assertEqual(list(worker.slots.values_list('id', flat=True)),
                         list(queue.slots.values_list('id', flat=True)))

        # worker has no job splits assigned to it
        self.assertEqual(worker.splits.count(), 0)

        self.assertEqual(queue.availability(), qsetup.CORES)
        self.assertEqual(queue.number_of_slots(), qsetup.CORES)
        self.assertEqual(queue.worker_availability(), [worker])

        # checks the single slot and priority
        slot = queue.slots.get()
        self.assertEqual(slot.quantity, qsetup.CORES)
        self.assertEqual(slot.priority, 0)
        self.assertEqual(slot.worker, worker)

        # checks no orphan slots exist
        self.assertEqual(Slot.objects.filter(queue=None).count(), 0)
        self.assertEqual(Slot.objects.filter(worker=None).count(), 0)


    def test_setup(self):

        self.check_default_config()


    def test_cmd_reset(self):

        # installs the default configuration command
        management.call_command('qsetup', verbosity=0, reset=True)
        self.check_default_config()


    def check_noprior_config(self):

        qs = Queue.objects.all()

        self.assertEqual(qs.count(), 3)

        q1, q2, q3 = qs.order_by('name')

        self.assertEqual(q1.name, 'q1')
        self.assertEqual(q2.name, 'q2')
        self.assertEqual(q3.name, 'q4')
        self.assertEqual(q1.splits().count(), 0)
        self.assertEqual(q2.splits().count(), 0)
        self.assertEqual(q3.splits().count(), 0)

        self.assertEqual(q1.number_of_slots(), 4)
        self.assertEqual(q2.number_of_slots(), 2)
        self.assertEqual(q3.number_of_slots(), 1)
        self.assertEqual(q1.availability(), 4)
        self.assertEqual(q2.availability(), 2)
        self.assertEqual(q3.availability(), 1)
        self.assertEqual(q1.environments.count(), 1)
        self.assertEqual(q2.environments.count(), 1)
        self.assertEqual(q3.environments.count(), 1)

        self.assertEqual(q1.environments.first(), q2.environments.first())
        self.assertEqual(q2.environments.first(), q3.environments.first())

        env = q1.environments.first()

        self.assertEqual(env.name, 'Python 2.7')
        self.assertEqual(env.version, '1.3.0')

        self.assertEqual(q1.slots.count(), 1)
        self.assertEqual(q2.slots.count(), 1)
        self.assertEqual(q3.slots.count(), 1)

        slot1 = q1.slots.first()
        slot2 = q2.slots.first()
        slot3 = q3.slots.first()

        self.assertEqual(slot1.quantity, 4)
        self.assertEqual(slot1.priority, 0)
        self.assertEqual(slot1.queue, q1)
        self.assertEqual(slot2.quantity, 2)
        self.assertEqual(slot2.priority, 0)
        self.assertEqual(slot2.queue, q2)
        self.assertEqual(slot3.quantity, 1)
        self.assertEqual(slot3.priority, 0)
        self.assertEqual(slot3.queue, q3)

        worker1 = slot1.worker
        worker2 = slot2.worker
        worker3 = slot3.worker

        self.assertEqual(worker1, worker2)
        self.assertEqual(worker2, worker3)

        self.assertEqual(worker1.name, 'node1')
        self.assertEqual(list(worker1.splits.all()), [])
        self.assertEqual(worker1.memory, 16*1024)
        self.assertEqual(worker1.cores, 4)
        self.assertEqual(worker1.available_cores(), 4)

        self.assertEqual(worker1.slots.count(), 3)

        self.assertEqual(set(worker1.slots.all()),
                         set(list(q1.slots.all()) + list(q2.slots.all()) + \
                             list(q3.slots.all())))

        avail1 = q1.worker_availability()
        self.assertEqual(avail1, [worker1])

        avail2 = q2.worker_availability()
        self.assertEqual(avail2, [worker1])

        avail3 = q3.worker_availability()
        self.assertEqual(avail2, [worker1])

        # checks no orphan slots exist
        self.assertEqual(Slot.objects.filter(queue=None).count(), 0)
        self.assertEqual(Slot.objects.filter(worker=None).count(), 0)


    def test_reconfigure_noprior(self):

        setup_backend(QUEUES_WITHOUT_PRIORITY)
        Worker.objects.update(active=True)
        self.check_noprior_config()


    def test_reconfigure_fail_qenv_used(self):

        fullname = 'user/user/single/1/single'
        xp = Experiment.objects.get(name=fullname.split('/')[-1])

        # schedules the experiment and check it
        xp.schedule()
        self.check_single(xp)

        try:
            setup_backend(QUEUES_WITHOUT_PRIORITY)
        except RuntimeError as e:
            assert str(e).find('on the following queue/environment combinations') != -1
        else:
            assert False, 'Queue re-configuration worked with q/env in use'


    def check_prior_config(self):

        qs = Queue.objects.all()

        self.assertEqual(qs.count(), 3)

        q1, q1_special, q2 = qs.order_by('name')

        self.assertEqual(q1.name, 'q1')
        self.assertEqual(q2.name, 'q2')
        self.assertEqual(q1_special.name, 'q1_special')
        self.assertEqual(q1.splits().count(), 0)
        self.assertEqual(q2.splits().count(), 0)
        self.assertEqual(q1_special.splits().count(), 0)

        self.assertEqual(q1.number_of_slots(), 8)
        self.assertEqual(q2.number_of_slots(), 4)
        self.assertEqual(q1_special.number_of_slots(), 8)
        self.assertEqual(q1.availability(), 8)
        self.assertEqual(q2.availability(), 4)
        self.assertEqual(q1_special.availability(), 8)
        self.assertEqual(q1.environments.count(), 1)
        self.assertEqual(q2.environments.count(), 1)
        self.assertEqual(q1_special.environments.count(), 1)

        self.assertEqual(q1.environments.first(), q2.environments.first())
        self.assertEqual(q2.environments.first(),
                         q1_special.environments.first())

        env = q1.environments.first()

        self.assertEqual(env.name, 'Python 2.7')
        self.assertEqual(env.version, '1.3.0')

        self.assertEqual(q1.slots.count(), 2)
        self.assertEqual(q1_special.slots.count(), 2)
        self.assertEqual(q2.slots.count(), 2)

        slot11, slot12 = q1.slots.all()
        slot1_special1, slot1_special2 = q1_special.slots.all()
        slot21, slot22 = q2.slots.all()

        self.assertEqual(slot11.quantity, 4)
        self.assertEqual(slot11.priority, 5)
        self.assertEqual(slot12.quantity, 4)
        self.assertEqual(slot12.priority, 0)
        self.assertEqual(slot11.queue, q1)
        self.assertEqual(slot12.queue, q1)

        self.assertEqual(slot21.quantity, 2)
        self.assertEqual(slot21.priority, 0)
        self.assertEqual(slot22.quantity, 2)
        self.assertEqual(slot22.priority, 10)
        self.assertEqual(slot21.queue, q2)
        self.assertEqual(slot22.queue, q2)

        self.assertEqual(slot1_special1.quantity, 4)
        self.assertEqual(slot1_special1.priority, 0)
        self.assertEqual(slot1_special2.quantity, 4)
        self.assertEqual(slot1_special2.priority, 5)
        self.assertEqual(slot1_special1.queue, q1_special)
        self.assertEqual(slot1_special2.queue, q1_special)

        worker1 = slot11.worker
        worker2 = slot12.worker
        worker21 = slot21.worker
        worker22 = slot22.worker
        worker1_special1 = slot1_special1.worker
        worker1_special2 = slot1_special2.worker

        self.assertEqual(worker1, worker21)
        self.assertEqual(worker1, worker1_special1)
        self.assertEqual(worker2, worker22)
        self.assertEqual(worker2, worker1_special2)

        self.assertEqual(worker1.name, 'node1')
        self.assertEqual(worker1.splits.count(), 0)
        self.assertEqual(worker1.memory, 32*1024)
        self.assertEqual(worker1.cores, 4)
        self.assertEqual(worker1.available_cores(), 4)

        self.assertEqual(worker2.name, 'node2')
        self.assertEqual(worker2.splits.count(), 0)
        self.assertEqual(worker2.memory, 16*1024)
        self.assertEqual(worker2.cores, 4)
        self.assertEqual(worker2.available_cores(), 4)

        self.assertEqual(worker1.slots.count(), 3)
        self.assertEqual(worker2.slots.count(), 3)

        avail1 = q1.worker_availability()
        self.assertEqual(avail1, [worker1, worker2])

        avail2 = q2.worker_availability()
        self.assertEqual(avail2, [worker2, worker1])

        avail1_special = q1_special.worker_availability()
        self.assertEqual(avail1_special, [worker2, worker1])

        # checks no orphan slots exist
        self.assertEqual(Slot.objects.filter(queue=None).count(), 0)
        self.assertEqual(Slot.objects.filter(worker=None).count(), 0)


    def test_reconfigure_priors(self):

        setup_backend(PRIORITY_QUEUES)
        Worker.objects.update(active=True)
        self.check_prior_config()
